import { Observable, pipe, BehaviorSubject, Subject, queueScheduler, Subscription, isObservable, EMPTY, identity, from, of } from 'rxjs';
import { filter, map, distinctUntilChanged, catchError, observeOn, withLatestFrom, tap, finalize, mergeMap } from 'rxjs/operators';

class StoreExtension {
  constructor() {
    this.sortOrder = 0
    /* DEFAULT */
    ;
  }

}
class Actions extends Observable {}
const EFFECT_METADATA_KEY = '@mini-rx/effectMetaData';

const miniRxNameSpace = '@mini-rx';

function createMiniRxActionType(miniRxActionType, featureKey) {
  return miniRxNameSpace + (featureKey ? '/' + featureKey : '') + '/' + miniRxActionType;
}
function createMiniRxAction(miniRxActionType, featureKey) {
  return {
    type: createMiniRxActionType(miniRxActionType, featureKey)
  };
}
const setStateActionTypeKey = 'setStateActionType'; // Type predicate

function isFeatureStoreSetStateAction(action) {
  return action[setStateActionTypeKey] === "@mini-rx/feature-store"
  /* FEATURE_STORE */
  ;
}
function isComponentStoreSetStateAction(action) {
  return action[setStateActionTypeKey] === "@mini-rx/component-store"
  /* COMPONENT_STORE */
  ;
}
const UNDO_ACTION = miniRxNameSpace + '/undo';
function undo(action) {
  return {
    type: UNDO_ACTION,
    payload: action
  };
}

function ofType(...allowedTypes) {
  return filter(action => allowedTypes.some(type => {
    return type === action.type;
  }));
}
function select(mapFn) {
  return pipe(map(mapFn), distinctUntilChanged());
}
function miniRxError(message) {
  throw new Error(miniRxNameSpace + ': ' + message);
}
function miniRxConsoleError(message, err) {
  console.error(miniRxNameSpace + ': ' + message + '\nDetails:', err);
}
function hasEffectMetaData(param) {
  // eslint-disable-next-line no-prototype-builtins
  return param.hasOwnProperty(EFFECT_METADATA_KEY);
} // Calculate the setState callback and remove internal properties (only display type and payload in the LoggingExtension and Redux DevTools)

function beautifyActionForLogging(action, state) {
  if (isFeatureStoreSetStateAction(action)) {
    const featureState = state[action.featureKey];
    return {
      type: action.type,
      payload: calcNewPartialState(featureState, action.stateOrCallback)
    };
  } else if (isComponentStoreSetStateAction(action)) {
    return {
      type: action.type,
      payload: calcNewPartialState(state, action.stateOrCallback)
    };
  }

  return action;
}

function calcNewPartialState(state, stateOrCallback) {
  return typeof stateOrCallback === 'function' ? stateOrCallback(state) : stateOrCallback;
}

function calcNewState(state, stateOrCallback) {
  const newPartialState = calcNewPartialState(state, stateOrCallback);
  return Object.assign(Object.assign({}, state), newPartialState);
}
function combineMetaReducers(metaReducers) {
  return reducer => {
    return metaReducers.reduceRight((previousValue, currentValue) => {
      return currentValue(previousValue);
    }, reducer);
  };
}
function sortExtensions(extensions) {
  return [...extensions].sort((a, b) => {
    return a.sortOrder - b.sortOrder;
  });
}

// Credits go to NgRx

function defaultEffectsErrorHandler(observable$, retryAttemptLeft = 10) {
  return observable$.pipe(catchError(error => {
    miniRxConsoleError(`An error occurred in the Effect. MiniRx resubscribed the Effect automatically and will do so ${retryAttemptLeft - 1} more times.\nPlease provide error handling inside the Effect using \`catchError\` or \`tapResponse\`.`, error);

    if (retryAttemptLeft <= 1) {
      return observable$; // last attempt
    } // Return observable that produces this particular effect


    return defaultEffectsErrorHandler(observable$, retryAttemptLeft - 1);
  }));
}

// Credits go to NgRx
// Copied from with small modifications: https://github.com/ngrx/platform/blob/12.0.0/modules/store/src/utils.ts
function combineReducers(reducers) {
  const reducerKeys = Object.keys(reducers);
  const reducerKeyLength = reducerKeys.length;
  return (state = {}, action) => {
    const stateKeysLength = Object.keys(state).length;
    let hasChanged = stateKeysLength !== reducerKeyLength;
    const nextState = {};

    for (let i = 0; i < reducerKeyLength; i++) {
      const key = reducerKeys[i];
      const reducer = reducers[key];
      const previousStateForKey = state[key];
      const nextStateForKey = reducer(previousStateForKey, action);
      nextState[key] = nextStateForKey;
      hasChanged = hasChanged || nextStateForKey !== previousStateForKey;
    }

    return hasChanged ? nextState : state;
  };
}

class State {
  constructor(initialState) {
    this.initialState = initialState;
    this.stateSource = new BehaviorSubject(this.initialState);
    this.state$ = this.stateSource.asObservable().pipe( // Skip the first (undefined) value of the BehaviorSubject
    // Very similar to a ReplaySubject(1), but more lightweight
    // Emits a state object (when calling the `set` method)
    filter(v => !!v));
  }

  get() {
    return this.stateSource.value;
  }

  set(v) {
    this.stateSource.next(v);
  }

  patch(stateOrCallback) {
    if (!this.get()) {
      throw new Error('State is not initialized.');
    }

    this.stateSource.next(calcNewState(this.get(), stateOrCallback));
  }

  select(mapFn) {
    if (!mapFn) {
      return this.state$;
    }

    return this.state$.pipe(select(mapFn));
  }

}

class ActionsOnQueue {
  constructor() {
    this.actionsSource = new Subject();
    this.actions$ = this.actionsSource.asObservable().pipe(observeOn(queueScheduler) // Prevent stack overflow: https://blog.cloudboost.io/so-how-does-rx-js-queuescheduler-actually-work-188c1b46526e
    );
  }

  dispatch(action) {
    this.actionsSource.next(action);
  }

}

let hasUndoExtension = false;
let isStoreInitialized = false; // REDUCER STATE
// exported for testing purposes

const reducerState = new State();
const reducer$ = reducerState.select(v => {
  const combinedMetaReducer = combineMetaReducers(v.metaReducers);
  const combinedReducer = v.combineReducersFn(v.featureReducers);
  return combinedMetaReducer(combinedReducer);
});

function hasFeatureReducers() {
  return !!Object.keys(reducerState.get().featureReducers).length;
}

function checkFeatureExists(featureKey) {
  if (reducerState.get().featureReducers.hasOwnProperty(featureKey)) {
    miniRxError(`Feature "${featureKey}" already exists.`);
  }
}

function addReducer(featureKey, reducer) {
  initStore();
  checkFeatureExists(featureKey);
  reducerState.patch(state => ({
    featureReducers: Object.assign(Object.assign({}, state.featureReducers), {
      [featureKey]: reducer
    })
  }));
}

function removeReducer(featureKey) {
  reducerState.patch(state => ({
    featureReducers: omit(state.featureReducers, featureKey)
  }));
} // exported for testing purposes


function addMetaReducers(...reducers) {
  reducerState.patch(state => ({
    metaReducers: [...state.metaReducers, ...reducers]
  }));
}

function updateCombineReducersFn(combineReducersFn) {
  reducerState.patch({
    combineReducersFn
  });
} // ACTIONS


const actionsOnQueue = new ActionsOnQueue();
const actions$ = actionsOnQueue.actions$; // APP STATE

const appState = new State(); // Wire up the Redux Store: Init reducer state, subscribe to the actions and reducer Observable
// Called by `configureStore` and `addReducer`

function initStore() {
  if (isStoreInitialized) {
    return;
  }

  reducerState.set({
    featureReducers: {},
    metaReducers: [],
    combineReducersFn: combineReducers
  });
  let reducer; // 👇 We could use `withLatestFrom` inside actionsOnQueue.actions$.pipe, but fewer operators = less bundle-size :)

  reducer$.subscribe(v => reducer = v); // Listen to the Actions stream and update state accordingly

  actionsOnQueue.actions$.subscribe(action => {
    const newState = reducer(appState.get(), // Initially undefined, but the reducer can handle undefined (by falling back to initial state)
    action);
    appState.set(newState);
  });
  isStoreInitialized = true;
}

function configureStore$1(config = {}) {
  var _a, _b;

  initStore();

  if (hasFeatureReducers()) {
    miniRxError('`configureStore` detected reducers. Did you instantiate FeatureStores before calling `configureStore`?');
  }

  if (config.combineReducersFn) {
    updateCombineReducersFn(config.combineReducersFn);
  }

  if ((_a = config.metaReducers) === null || _a === void 0 ? void 0 : _a.length) {
    addMetaReducers(...config.metaReducers);
  }

  if ((_b = config.extensions) === null || _b === void 0 ? void 0 : _b.length) {
    const sortedExtensions = sortExtensions(config.extensions);
    sortedExtensions.forEach(extension => addExtension(extension));
  }

  if (config.reducers) {
    Object.keys(config.reducers).forEach(featureKey => {
      addReducer(featureKey, config.reducers[featureKey]); // config.reducers! (prevent TS2532: Object is possibly 'undefined')
    });
  }

  if (config.initialState) {
    appState.set(config.initialState);
  }

  dispatch(createMiniRxAction("init"
  /* INIT */
  ));
}
function addFeature(featureKey, reducer, config = {}) {
  var _a;

  reducer = ((_a = config.metaReducers) === null || _a === void 0 ? void 0 : _a.length) ? combineMetaReducers(config.metaReducers)(reducer) : reducer;

  if (typeof config.initialState !== 'undefined') {
    reducer = createReducerWithInitialState(reducer, config.initialState);
  }

  addReducer(featureKey, reducer);
  dispatch(createMiniRxAction("init"
  /* INIT */
  , featureKey));
}
function removeFeature(featureKey) {
  removeReducer(featureKey);
  dispatch(createMiniRxAction("destroy"
  /* DESTROY */
  , featureKey));
}
function effect(effect$) {
  const effectWithErrorHandler$ = defaultEffectsErrorHandler(effect$);
  effectWithErrorHandler$.subscribe(action => {
    let shouldDispatch = true;

    if (hasEffectMetaData(effect$)) {
      const metaData = effect$[EFFECT_METADATA_KEY];
      shouldDispatch = !!metaData.dispatch;
    }

    if (shouldDispatch) {
      dispatch(action);
    }
  });
} // exported for testing purposes

function addExtension(extension) {
  const metaReducer = extension.init();

  if (metaReducer) {
    addMetaReducers(metaReducer);
  }

  if (extension.id === 1
  /* UNDO */
  ) {
    hasUndoExtension = true;
  }
}
function dispatch(action) {
  actionsOnQueue.dispatch(action);
}

function createReducerWithInitialState(reducer, initialState) {
  return (state = initialState, action) => {
    return reducer(state, action);
  };
}

function omit(object, keyToOmit) {
  return Object.keys(object).filter(key => key !== keyToOmit).reduce((prevValue, key) => {
    prevValue[key] = object[key];
    return prevValue;
  }, {});
}

var storeCore = /*#__PURE__*/Object.freeze({
    __proto__: null,
    get hasUndoExtension () { return hasUndoExtension; },
    reducerState: reducerState,
    addMetaReducers: addMetaReducers,
    actions$: actions$,
    appState: appState,
    configureStore: configureStore$1,
    addFeature: addFeature,
    removeFeature: removeFeature,
    effect: effect,
    addExtension: addExtension,
    dispatch: dispatch
});

class Store {}
let isStoreConfigured = false;
function configureStore(config) {
  if (!isStoreConfigured) {
    configureStore$1(config);

    isStoreConfigured = true;
    return {
      feature: addFeature,
      select: appState.select.bind(appState),
      dispatch,
      effect
    };
  }

  miniRxError('`configureStore` was called multiple times.');
}

class BaseStore {
  constructor() {
    /**
     * @internal Used by ComponentStore/FeatureStore
     */
    this._sub = new Subscription();
    /**
     * @internal Used by ComponentStore/FeatureStore
     */

    this._state = new State();
    /** @deprecated Use the `select` method without arguments */

    this.state$ = this._state.select();
    this.isStateInitialized = false;
    this.notInitializedErrorMessage = `${this.constructor.name} has no initialState yet. ` + `Please provide an initialState before updating/getting state.`;
    this.initializedErrorMessage = `${this.constructor.name} has initialState already.`;
    this.select = this._state.select.bind(this._state);
  }

  get state() {
    this.assertStateIsInitialized();
    return this._state.get();
  } // Called by ComponentStore/FeatureStore


  setInitialState(initialState) {
    this.assertStateIsNotInitialized();
    this.isStateInitialized = true; // Update state happens in ComponentStore/FeatureStore
  }

  setState(stateOrCallback, name) {
    const dispatchFn = (stateOrCallback, name) => {
      this.assertStateIsInitialized();
      return this._dispatchSetStateAction(stateOrCallback, name);
    };

    return isObservable(stateOrCallback) ? this._sub.add(stateOrCallback.subscribe(v => dispatchFn(v, name))) : dispatchFn(stateOrCallback, name);
  }

  effect(effectFn) {
    const subject = new Subject();
    const effect$ = effectFn(subject);
    const effectWithDefaultErrorHandler = defaultEffectsErrorHandler(effect$);

    this._sub.add(effectWithDefaultErrorHandler.subscribe());

    return observableOrValue => {
      isObservable(observableOrValue) ? this._sub.add(observableOrValue.subscribe(v => subject.next(v))) : subject.next(observableOrValue);
    };
  }

  assertStateIsInitialized() {
    if (!this.isStateInitialized) {
      miniRxError(this.notInitializedErrorMessage);
    }
  }

  assertStateIsNotInitialized() {
    if (this.isStateInitialized) {
      miniRxError(this.initializedErrorMessage);
    }
  }

  destroy() {
    this._sub.unsubscribe();
  }
  /**
   * @internal
   * Can be called by Angular if ComponentStore/FeatureStore is provided in a component
   */


  ngOnDestroy() {
    this.destroy();
  }

}

class FeatureStore extends BaseStore {
  constructor(featureKey, initialState, config = {}) {
    super();
    this.featureId = generateId();
    this._featureKey = config.multi ? featureKey + '-' + generateId() : featureKey;

    if (initialState) {
      this.setInitialState(initialState);
    }
  }

  get featureKey() {
    return this._featureKey;
  }

  setInitialState(initialState) {
    super.setInitialState(initialState);
    addFeature(this._featureKey, createFeatureStoreReducer(this.featureId, initialState));

    this._sub.add(appState.select(state => state[this.featureKey]).subscribe(v => this._state.set(v)));
  }
  /** @internal
   * Implementation of abstract method from BaseStore
   */


  _dispatchSetStateAction(stateOrCallback, name) {
    const action = createSetStateAction$1(stateOrCallback, this.featureId, this.featureKey, name);
    dispatch(action);
    return action;
  } // Implementation of abstract method from BaseStore


  undo(action) {
    hasUndoExtension ? dispatch(undo(action)) : miniRxError('UndoExtension is not initialized.');
  }

  destroy() {
    super.destroy();
    removeFeature(this._featureKey);
  }

}

function createFeatureStoreReducer(featureId, initialState) {
  return (state = initialState, action) => {
    if (isFeatureStoreSetStateAction(action) && action.featureId === featureId) {
      return calcNewState(state, action.stateOrCallback);
    }

    return state;
  };
}

function createSetStateAction$1(stateOrCallback, featureId, featureKey, name) {
  const miniRxActionType = "set-state"
  /* SET_STATE */
  ;
  return {
    setStateActionType: "@mini-rx/feature-store"
    /* FEATURE_STORE */
    ,
    type: createMiniRxActionType(miniRxActionType, featureKey) + (name ? '/' + name : ''),
    stateOrCallback,
    featureId,
    featureKey
  };
} // Simple alpha numeric ID: https://stackoverflow.com/a/12502559/453959
// This isn't a real GUID!


function generateId() {
  return Math.random().toString(36).slice(2);
}

function createFeatureStore(featureKey, initialState, config = {}) {
  return new FeatureStore(featureKey, initialState, config);
}

let componentStoreConfig = undefined;
function configureComponentStores(config) {
  if (!componentStoreConfig) {
    componentStoreConfig = config;
    return;
  }

  miniRxError('`configureComponentStores` was called multiple times.');
}
const csFeatureKey = 'component-store';
class ComponentStore extends BaseStore {
  constructor(initialState, config) {
    super();
    this.actionsOnQueue = new ActionsOnQueue();
    this.hasUndoExtension = false;
    this.extensions = []; // This is a class property just for testing purposes

    const metaReducers = [];

    if (config === null || config === void 0 ? void 0 : config.extensions) {
      if (config.extensions && (componentStoreConfig === null || componentStoreConfig === void 0 ? void 0 : componentStoreConfig.extensions)) {
        this.extensions = mergeExtensions(componentStoreConfig.extensions, config.extensions);
      } else {
        this.extensions = config.extensions;
      }
    } else if (componentStoreConfig === null || componentStoreConfig === void 0 ? void 0 : componentStoreConfig.extensions) {
      this.extensions = componentStoreConfig.extensions;
    }

    sortExtensions(this.extensions).forEach(ext => {
      if (!ext.hasCsSupport) {
        miniRxError(`Extension "${ext.constructor.name}" is not supported by Component Store.`);
      }

      metaReducers.push(ext.init()); // Non-null assertion: Here we know for sure: init will return a MetaReducer

      if (ext.id === 1
      /* UNDO */
      ) {
        this.hasUndoExtension = true;
      }
    });
    this.combinedMetaReducer = combineMetaReducers(metaReducers);

    this._sub.add(this.actionsOnQueue.actions$.subscribe(action => {
      const newState = this.reducer( // We are sure, there is a reducer!
      this._state.get(), // Initially undefined, but the reducer can handle undefined (by falling back to initial state)
      action);

      this._state.set(newState);
    }));

    if (initialState) {
      this.setInitialState(initialState);
    }
  }

  setInitialState(initialState) {
    super.setInitialState(initialState);
    this.reducer = this.combinedMetaReducer(createComponentStoreReducer(initialState));
    this.dispatch(createMiniRxAction("init"
    /* INIT */
    , csFeatureKey));
  }
  /** @internal
   * Implementation of abstract method from BaseStore
   */


  _dispatchSetStateAction(stateOrCallback, name) {
    const action = createSetStateAction(stateOrCallback, name);
    this.dispatch(action);
    return action;
  }

  dispatch(action) {
    this.actionsOnQueue.dispatch(action);
  } // Implementation of abstract method from BaseStore


  undo(action) {
    this.hasUndoExtension ? this.dispatch(undo(action)) : miniRxError(`${this.constructor.name} has no UndoExtension yet.`);
  }

  destroy() {
    if (this.reducer) {
      // Dispatch an action really just for logging via LoggerExtension
      // Only dispatch if a reducer exists (if an initial state was provided or setInitialState was called)
      this.dispatch(createMiniRxAction("destroy"
      /* DESTROY */
      , csFeatureKey));
    }

    super.destroy();
  }

}

function createSetStateAction(stateOrCallback, name) {
  const miniRxActionType = "set-state"
  /* SET_STATE */
  ;
  return {
    setStateActionType: "@mini-rx/component-store"
    /* COMPONENT_STORE */
    ,
    type: createMiniRxActionType(miniRxActionType, csFeatureKey) + (name ? '/' + name : ''),
    stateOrCallback
  };
}

function createComponentStoreReducer(initialState) {
  return (state = initialState, action) => {
    if (isComponentStoreSetStateAction(action)) {
      return calcNewState(state, action.stateOrCallback);
    }

    return state;
  };
}

function mergeExtensions(global, local) {
  // Local extensions overwrite the global extensions
  // If extension is global and local => use local
  // If extension is only global => use global
  // If extension is only local => use local
  const extensions = [];
  let globalCopy = [...global];
  let localCopy = [...local];
  global.forEach(globalExt => {
    local.forEach(localExt => {
      if (localExt.id === globalExt.id) {
        // Found extension which is global and local
        extensions.push(localExt); // Use local!

        localCopy = localCopy.filter(item => item.id !== localExt.id); // Remove found extension from local

        globalCopy = globalCopy.filter(item => item.id !== globalExt.id); // Remove found extension from global
      }
    });
  });
  return [...extensions, ...localCopy, ...globalCopy // Global only
  ];
}

function createComponentStore(initialState, config) {
  return new ComponentStore(initialState, config);
}

// Credits go to Alexander Reardon
// Copied from with small modifications: https://github.com/alexreardon/memoize-one/tree/v6.0.0/src
// MIT License
//
// Copyright (c) 2019 Alexander Reardon
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
function isEqual(first, second) {
  if (first === second) {
    return true;
  } // Special case for NaN (NaN !== NaN)


  if (Number.isNaN(first) && Number.isNaN(second)) {
    return true;
  }

  return false;
}

function areInputsEqual(newInputs, lastInputs) {
  // no checks needed if the inputs length has changed
  if (newInputs.length !== lastInputs.length) {
    return false;
  } // Using for loop for speed. It generally performs better than array.every
  // https://github.com/alexreardon/memoize-one/pull/59


  for (let i = 0; i < newInputs.length; i++) {
    if (!isEqual(newInputs[i], lastInputs[i])) {
      return false;
    }
  }

  return true;
}

function memoizeOne(resultFn) {
  let cache = null; // breaking cache when context (this) or arguments change

  function memoized(...newArgs) {
    if (cache && cache.lastThis === this && areInputsEqual(newArgs, cache.lastArgs)) {
      return cache.lastResult;
    } // Throwing during an assignment aborts the assignment: https://codepen.io/alexreardon/pen/RYKoaz
    // Doing the lastResult assignment first so that if it throws
    // the cache will not be overwritten


    const lastResult = resultFn.apply(this, newArgs);
    cache = {
      lastResult,
      lastArgs: newArgs,
      lastThis: this
    };
    return lastResult;
  }

  return memoized;
}

// Credits go to NgRx
function createSelector(...args) {
  const selectors = args.slice(0, args.length - 1);
  const projector = args[args.length - 1];
  const memoizedProjector = memoizeOne(projector);
  return memoizeOne(state => {
    const selectorResults = selectors.map(fn => fn(state));
    return memoizedProjector.apply(null, selectorResults);
  });
}
/** @deprecated Use `createFeatureStateSelector` which is more in line with `createComponentStateSelector` */

function createFeatureSelector(featureKey) {
  if (featureKey) {
    return createSelector(state => state[featureKey], featureState => featureState);
  }

  return state => state; // Do not memoize: when used with FeatureStore there is a new state object created for every `setState`
}
const createFeatureStateSelector = createFeatureSelector;
function createComponentStateSelector() {
  return state => state;
}

const defaultOptions = {
  name: 'MiniRx - Redux DevTools',
  traceLimit: 25
};
class ReduxDevtoolsExtension extends StoreExtension {
  constructor(options) {
    super();
    this.options = options;
    this.id = 3
    /* REDUX_DEVTOOLS */
    ;

    if (!window) {
      miniRxError('The Redux DevTools are only supported in browser environments.');
    }

    this._optionsForNgExtension = options;
    this.devtoolsExtension = window.__REDUX_DEVTOOLS_EXTENSION__;
    this.options = Object.assign(Object.assign({}, defaultOptions), this.options);
  }

  get optionsForNgExtension() {
    return this._optionsForNgExtension;
  }

  init() {
    if (this.devtoolsExtension) {
      this.devtoolsConnection = this.devtoolsExtension.connect(this.options);
      actions$.pipe(withLatestFrom(appState.select()), tap(([action, state]) => {
        const actionForDevTools = beautifyActionForLogging(action, state);
        this.devtoolsConnection.send(actionForDevTools, state);
      })).subscribe();
      this.devtoolsConnection.subscribe(this.onDevToolsMessage.bind(this));
    }
  }

  onDevToolsMessage(message) {
    if (message.type === "DISPATCH"
    /* DISPATCH */
    ) {
      switch (message.payload.type) {
        case "JUMP_TO_STATE"
        /* JUMP_TO_STATE */
        :
        case "JUMP_TO_ACTION"
        /* JUMP_TO_ACTION */
        :
          this.updateState(JSON.parse(message.state));
      }
    }
  }

  updateState(state) {
    appState.set(state);
  }

}

class LoggerExtension extends StoreExtension {
  constructor() {
    super(...arguments);
    this.id = 2
    /* LOGGER */
    ;
    this.hasCsSupport = true;
  }

  init() {
    return loggerMetaReducer;
  }

}

function loggerMetaReducer(reducer) {
  return (state, action) => {
    const actionToLog = beautifyActionForLogging(action, state);
    const nextState = reducer(state, action);
    console.log('%c' + action.type, 'color: #25c2a0', '\nAction:', actionToLog, '\nState: ', nextState);
    return nextState;
  };
}

// Copied from with small modifications: https://github.com/jsdf/deep-freeze/blob/v1.1.1/index.js
// Original License:
// This software is released to the public domain.
//
// It is based in part on the deepFreeze function from:
// https://developer.mozilla.org/en-US/docs/JavaScript/Reference/Global_Objects/Object/freeze
//
// https://developer.mozilla.org/en-US/docs/Project:Copyrights
function deepFreeze(o) {
  Object.freeze(o);
  const oIsFunction = typeof o === 'function';
  const hasOwnProp = Object.prototype.hasOwnProperty;
  Object.getOwnPropertyNames(o).forEach(function (prop) {
    if (hasOwnProp.call(o, prop) && (oIsFunction ? prop !== 'caller' && prop !== 'callee' && prop !== 'arguments' : true) && o[prop] !== null && (typeof o[prop] === 'object' || typeof o[prop] === 'function') && !Object.isFrozen(o[prop])) {
      deepFreeze(o[prop]);
    }
  });
  return o;
}

// Credits go to Brandon Roberts
class ImmutableStateExtension extends StoreExtension {
  constructor() {
    super(...arguments);
    this.id = 0
    /* IMMUTABLE_STATE */
    ;
    this.hasCsSupport = true;
  }

  init() {
    return storeFreeze;
  }

}

function storeFreeze(reducer) {
  return (state, action) => {
    if (state) {
      deepFreeze(state);
    }

    const nextState = reducer(state, action);
    deepFreeze(nextState);
    return nextState;
  };
}

// Credits go to Brecht Billiet
const defaultBufferSize = 100;
class UndoExtension extends StoreExtension {
  constructor(config = {
    bufferSize: defaultBufferSize
  }) {
    super();
    this.config = config;
    this.id = 1
    /* UNDO */
    ;
    this.sortOrder = 1
    /* UNDO_EXTENSION */
    ;
    this.hasCsSupport = true;
  }

  init() {
    return createUndoMetaReducer(this.config.bufferSize);
  }

}

function createUndoMetaReducer(bufferSize) {
  let executedActions = [];
  let initialState;
  return rootReducer => {
    return (state, action) => {
      if (action.type === UNDO_ACTION) {
        // if the action is UNDO_ACTION,
        // then call all the actions again on the rootReducer,
        // except the one we want to rollback
        let newState = initialState;
        executedActions = executedActions.filter(eAct => eAct !== action.payload); // update the state for every action until we get the
        // exact same state as before, but without the action we want to rollback

        executedActions.forEach(executedAction => {
          newState = rootReducer(newState, executedAction);
        });
        return newState;
      } else {
        // push every action that isn't UNDO_ACTION
        executedActions.push(action);
      }

      const updatedState = rootReducer(state, action);

      if (executedActions.length === bufferSize + 1) {
        const firstAction = executedActions[0]; // calculate the state x (bufferSize) actions ago

        initialState = rootReducer(initialState, firstAction); // keep the correct actions

        executedActions = executedActions.slice(1, bufferSize + 1);
      }

      return updatedState;
    };
  };
}

// Credits go to NgRx
function tapResponse(fn1OrObject, fn2, fn3) {
  return source => {
    const tapResponseObj = typeof fn1OrObject === 'function' ? {
      next: fn1OrObject,
      error: fn2,
      finalize: fn3
    } : fn1OrObject;
    return source.pipe(tap({
      next: v => {
        if (tapResponseObj.next) {
          tryCatch(() => tapResponseObj.next(v), 'next');
        }
      },
      error: err => tryCatch(() => tapResponseObj.error(err), 'error')
    }), catchError(() => EMPTY), tapResponseObj.finalize ? finalize(() => {
      tryCatch(() => tapResponseObj.finalize(), 'finalize');
    }) : identity // Conditionally apply operator: https://rxjs.dev/api/index/function/identity
    );
  };
}

function tryCatch(fn, whichCallback) {
  try {
    fn();
  } catch (error) {
    miniRxConsoleError(`An error occurred in the \`tapResponse\` ${whichCallback} callback.`, error);
  }
}

function mapResponse(mapFn, errorFn) {
  return source => source.pipe(map(v => mapFn(v)), mergeMap(mapFnResult => {
    return Array.isArray(mapFnResult) ? from(mapFnResult) : of(mapFnResult);
  }), catchError(err => {
    const errorFnResult = errorFn(err);
    return errorFnResult ? Array.isArray(errorFnResult) ? from(errorFnResult) : of(errorFnResult) : EMPTY;
  }));
}

// Credits go to NgRx
const DEFAULT_EFFECT_CONFIG = {
  dispatch: true
};
function createEffect(v, config) {
  const value = Object.assign(Object.assign({}, DEFAULT_EFFECT_CONFIG), config);
  Object.defineProperty(v, EFFECT_METADATA_KEY, {
    value
  });
  return v;
}

export { Actions, ComponentStore, FeatureStore, ImmutableStateExtension, LoggerExtension, ReduxDevtoolsExtension, Store, StoreExtension, UndoExtension, storeCore as _StoreCore, actions$, configureComponentStores, configureStore, createComponentStateSelector, createComponentStore, createEffect, createFeatureSelector, createFeatureStateSelector, createFeatureStore, createSelector, hasEffectMetaData, mapResponse, ofType, tapResponse, undo };
